"""
Copyright 2021, CCL Forensics
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import pathlib
import sqlite3
from ccl_chromium_reader import ccl_chromium_sessionstorage

__version__ = "0.1"
__description__ = "Dumps a Chromium sessionstorage leveldb to sqlite for review"
__contact__ = "Alex Caithness"

DB_SCHEMA = """
CREATE TABLE "hosts" ("_id" INTEGER PRIMARY KEY AUTOINCREMENT, "host" TEXT);
CREATE TABLE "guids" ("_id" INTEGER PRIMARY KEY AUTOINCREMENT, "guid" TEXT);
CREATE TABLE "items" ("_id" INTEGER PRIMARY KEY AUTOINCREMENT, 
                      "host" INTEGER, 
                      "guid" INTEGER, 
                      "ldbseq" INTEGER, 
                      "key" TEXT, 
                      "value" TEXT);
CREATE INDEX "item_host" ON "items" ("host");  
CREATE INDEX "item_ldbseq" ON "items" ("ldbseq");

CREATE VIEW items_view AS 
    SELECT "items"."ldbseq",
      "hosts"."host",
      "items"."key",
      "items"."value",
      "guids"."guid"
    FROM "items"
      LEFT JOIN "hosts" ON "items"."host" = "hosts"."_id"
      LEFT JOIN "guids" ON "items"."guid" = "guids"."_id"
    ORDER BY "items"."ldbseq";
"""

INSERT_HOST_SQL = """INSERT INTO "hosts" ("host") VALUES (?);"""
INSERT_ITEM_SQL = """INSERT INTO "items" (host, guid, ldbseq, key, value) VALUES (?, ?, ?, ?, ?);"""


def main(args):
    level_db_in_dir = pathlib.Path(args[0])
    db_out_path = pathlib.Path(args[1])

    if db_out_path.exists():
        raise ValueError("output database already exists")

    session_storage = ccl_chromium_sessionstorage.SessionStoreDb(level_db_in_dir)
    out_db = sqlite3.connect(db_out_path)
    out_db.executescript(DB_SCHEMA)
    cur = out_db.cursor()
    for host in session_storage.iter_hosts():
        cur.execute(INSERT_HOST_SQL, (host,))
        cur.execute("SELECT last_insert_rowid();")
        host_id = cur.fetchone()[0]
        host_kvs = session_storage.get_all_for_host(host)

        for key, values in host_kvs.items():
            for value in values:
                cur.execute(INSERT_ITEM_SQL, (host_id, None, value.leveldb_sequence_number, key, value.value))

    for key, value in session_storage.iter_orphans():
        cur.execute(INSERT_ITEM_SQL, (None, None, value.leveldb_sequence_number, key, value.value))

    cur.close()
    out_db.commit()
    out_db.close()


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print(f"{pathlib.Path(sys.argv[0]).name} <leveldb dir> <out.db>")
        exit(1)
    main(sys.argv[1:])



"""
Copyright 2022, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import struct
import io
import typing

__version__ = "0.8"
__description__ = "Module for naive parsing of Protocol Buffers"
__contact__ = "Alex Caithness"

DEBUG = False


class Empty:
    value = None


class ProtoObject:
    def __init__(self, tag, name, value):
        self.tag = tag
        self.name = name
        self.value = value
        self.wire = tag & 0x07

    def __str__(self):
        if self.name:
            return "{0} ({1}): {2}".format(
                self.tag if self.tag > 0x7f else hex(self.tag), self.name, repr(self.value))
        else:
            return "{0}: {1}".format(
                self.tag if self.tag > 0x7f else hex(self.tag), repr(self.value))
    __repr__ = __str__

    @property
    def friendly_tag(self) -> int:
        """:return the "real" tag (i.e. the one that would be seen inside the .proto schema)"""
        return self.tag >> 3

    def get_items_by_tag(self, tag_id: int) -> list[typing.Any]:
        """
        :param tag_id: the tag id for the child items
        :return: list of child items with this tag number
        """
        if not isinstance(self.value, list):
            raise ValueError("This object does not support child items")
        if not isinstance(tag_id, int):
            raise TypeError("Expected type: int; actual type: {0}".format(type(tag_id)))
        return [x for x in self.value if x.tag == tag_id]

    def get_items_by_name(self, name: str) -> list[typing.Any]:
        """
        :param name: the field name for the child items
        :return: list of child items with this name
        """
        if not isinstance(self.value, list):
            raise ValueError("This object does not support child items")
        if not isinstance(name, str):
            raise TypeError("Expected type: str; actual type: {0}".format(type(name)))
        return [x for x in self.value if x.name == name]

    def only(self, name: str, default=Empty):
        """
        Returns a single item which matches the name parameter. Use this to streamline getting non-repeating items
        :param name: the name of the child item
        :param default: optional: the value to return if the item is not present (default: None)
        :return: the single child item
        :exception: ValueError: if there is more than one child item which matches this name
        """
        got = self.get_items_by_name(name)
        if len(got) == 0:
            return default
        elif len(got) == 1:
            return got[0]
        else:
            raise ValueError("More than one value with this key")

    def __getitem__(self, key: typing.Union[str, int]) -> list[typing.Any]:
        if isinstance(key, str):
            return self.get_items_by_name(key)
        elif isinstance(key, int):
            return self.get_items_by_tag(key)
        else:
            raise TypeError("Key should be int or str; actual type: {0}".format(type(key)))

    def __len__(self) -> int:
        return self.value.__len__()

    def __iter__(self):
        if not isinstance(self.value, list):
            raise ValueError("This object does not support child items")
        else:
            yield from (x.tag for x in self.value)


class ProtoDecoder:
    def __init__(self, object_name, func):
        self.func = func
        self.object_name = object_name

    def __call__(self, arg):
        return self.func(arg)


def _read_le_varint(stream: typing.BinaryIO, is_32bit=False) -> typing.Optional[typing.Tuple[int, bytes]]:
    # this only outputs unsigned
    limit = 5 if is_32bit else 10
    i = 0
    result = 0
    underlying_bytes = []
    while i < limit:  # 64 bit max possible?
        raw = stream.read(1)
        if len(raw) < 1:
            return None
        tmp, = raw
        underlying_bytes.append(tmp)
        result |= ((tmp & 0x7f) << (i * 7))
        if (tmp & 0x80) == 0:
            break
        i += 1
    return result, bytes(underlying_bytes)


def read_le_varint(stream: typing.BinaryIO, is_32bit=False) -> typing.Optional[int]:
    x = _read_le_varint(stream, is_32bit)
    if x is None:
        return None
    else:
        return x[0]


def read_le_varint32(stream: typing.BinaryIO) -> typing.Optional[int]:
    return read_le_varint(stream, True)


def read_tag(
        stream: typing.BinaryIO,
        tag_mappings: dict[int, typing.Callable[[typing.BinaryIO], typing.Any]],
        log_out=sys.stderr, use_friendly_tag=False) -> typing.Optional[ProtoObject]:
    tag_id = read_le_varint(stream)
    if tag_id is None: 
        return None
    decoder = tag_mappings.get(tag_id if not use_friendly_tag else tag_id >> 3)
    name = None
    if isinstance(decoder, ProtoDecoder):
        name = decoder.object_name
    
    available_wirebytes = io.BytesIO(_get_bytes_for_wiretype(tag_id, stream))

    tag_value = decoder(available_wirebytes) if decoder else _fallback_decode(
        tag_id, available_wirebytes, log_out)
    
    return ProtoObject(tag_id, name, tag_value)


def read_protobuff(
        stream: typing.BinaryIO,
        tag_mappings: dict [int, typing.Callable[[typing.BinaryIO], typing.Any]],
        use_friendly_tag=False) -> list[ProtoObject]:
    result = []
    while True:
        tag = read_tag(stream, tag_mappings, use_friendly_tag=use_friendly_tag)
        if tag is None:
            break
        result.append(tag)

    return result


def read_blob(stream: typing.BinaryIO) -> bytes:
    blob_length = read_le_varint(stream)
    blob = stream.read(blob_length)
    return blob


def read_string(stream: typing.BinaryIO) -> str:
    raw_string = read_blob(stream)
    string = raw_string.decode("utf-8")
    return string


def read_double(stream: typing.BinaryIO) -> float:
    return struct.unpack("<d", stream.read(8))[0]


def read_long(stream: typing.BinaryIO) -> int:
    return struct.unpack("<q", stream.read(8))[0]


def read_int(stream: typing.BinaryIO) -> int:
    return struct.unpack("<i", stream.read(4))[0]


def read_embedded_protobuf(stream: typing.BinaryIO, mappings, use_friendly_tag=False) -> list[ProtoObject]:
    blob_blob = read_blob(stream)
    blob_stream = io.BytesIO(blob_blob)
    return read_protobuff(blob_stream, mappings, use_friendly_tag)


def read_fixed_blob(stream: typing.BinaryIO, length: int) -> bytes:
    data = stream.read(length)
    if len(data) != length:
        raise ValueError("Couldn't read enough data")
    return data


_fallback_wire_types = {
    0: read_le_varint,
    1: lambda x: read_fixed_blob(x, 8),
    2: read_blob,
    5: lambda x: read_fixed_blob(x, 4)
    }

_wire_type_friendly_names = {
    0: "Varint",
    1: "64-Bit",
    2: "Length Delimited",
    5: "32-Bit"
    }


def _get_bytes_for_wiretype(tag_id: int, stream: typing.BinaryIO):
    wire_type = tag_id & 0x07
    if wire_type == 0:
        read_bytes = []
        for i in range(10):
            x = stream.read(1)[0]
            read_bytes.append(x)
            if x & 0x80 == 0:
                break
        buffer = bytes(read_bytes)
    elif wire_type == 1:
        buffer = stream.read(8)
    elif wire_type == 2:
        l, b = _read_le_varint(stream)
        available_bytes = stream.read(l)
        if len(available_bytes) < l:
            raise ValueError("Stream too short")
        buffer = b + available_bytes
    elif wire_type == 5:
        buffer = stream.read(4)
    else:
        raise ValueError("Invalid wiretype")

    return buffer


def _fallback_decode(tag_id, stream, log):
    fallback_func = _fallback_wire_types.get(tag_id & 0x07)
    if not fallback_func:
        raise ValueError("No appropriate fallback function for tag {0} (wire type {1})".format(
            tag_id, tag_id & 0x07))
    if DEBUG:
        log.write("Tag {0} ({1}) not defined, using fallback decoding.\n".format(
            tag_id if tag_id > 0x7f else hex(tag_id), _wire_type_friendly_names[tag_id & 0x07]))
    return fallback_func(stream)



"""
Copyright 2024, CCL Forensics
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import dataclasses
import datetime
import typing
import pathlib
import collections.abc as col_abc

import gzip
import zlib
import brotli

from . import ccl_chromium_localstorage
from . import ccl_chromium_sessionstorage
from . import ccl_chromium_indexeddb
from . import ccl_chromium_history
from . import ccl_chromium_cache
from . import ccl_shared_proto_db_downloads

from .common import KeySearch, is_keysearch_hit

__version__ = "0.4.1"
__description__ = "Module to consolidate and simplify access to data stores in the chrom(e|ium) profile folder"
__contact__ = "Alex Caithness"

# TODO: code currently assumes that lazy-loaded stores are present - they might not be, need some kind of guard object?
# TODO: paths are currently based around an assumption of a desktop layout. Mobile (and MacOS) will differ (e.g., the
#  cache location).

SESSION_STORAGE_FOLDER_PATH = pathlib.Path("Session Storage")
LOCAL_STORAGE_FOLDER_PATH = pathlib.Path("Local Storage", "leveldb")
INDEXEDDB_FOLDER_PATH = pathlib.Path("IndexedDB")
HISTORY_DB_PATH = pathlib.Path("History")
CACHE_PATH = pathlib.Path("Cache", "Cache_Data")
SHARED_PROTO_DB_FOLDER_PATH = pathlib.Path("shared_proto_db")


@dataclasses.dataclass(frozen=True, repr=False)
class CacheResult:
    key: ccl_chromium_cache.CacheKey = dataclasses.field(repr=True)
    metadata: ccl_chromium_cache.CachedMetadata
    data: bytes
    metadata_location: ccl_chromium_cache.CacheFileLocation
    data_location: ccl_chromium_cache.CacheFileLocation
    was_decompressed: bool
    duplicate_key_index: int


class ChromiumProfileFolder:
    """
    A class representing a Chrom(e|ium) profile folder with programmatic access to various different data stores.
    Where appropriate, resources are loaded on demand.
    """

    def __init__(self, path: pathlib.Path, *, cache_folder: typing.Optional[pathlib.Path]=None):
        """
        Constructor

        :param path: Path to the profile folder (usually named Default, Profile 1, Profile 2, etc.)
        :param cache_folder: optionally a path to a cache folder, for platforms (such as Android) which
                             place the cache data outside the profile folder.
        """
        if not path.is_dir():
            raise NotADirectoryError(f"Could not find the folder: {path}")

        self._path = path

        if cache_folder is not None and not cache_folder.is_dir():
            raise NotADirectoryError(f"Could not find the folder: {cache_folder}")

        self._external_cache_folder = cache_folder

        # Data stores are populated lazily where appropriate
        # Webstorage
        self._local_storage: typing.Optional[ccl_chromium_localstorage.LocalStoreDb] = None
        self._session_storage: typing.Optional[ccl_chromium_sessionstorage.SessionStoreDb] = None

        # IndexedDb
        # Dictionary which when first populated will initially contain the domains as keys with None as the value for
        #  each. This will be initially populated on demand.
        self._indexeddb_databases: typing.Optional[dict[str, typing.Optional[ccl_chromium_indexeddb.WrappedIndexDB]]] = None
        self._lazy_populate_indexeddb_list()

        # History
        self._history: typing.Optional[ccl_chromium_history.HistoryDatabase] = None

        # Cache
        self._cache: typing.Optional[ccl_chromium_cache.ChromiumCache] = None

    def close(self):
        """
        Closes any resources currently open in this profile folder.
        """
        if self._local_storage is not None:
            self._local_storage.close()
        if self._session_storage is not None:
            self._session_storage.close()
        for idb in self._indexeddb_databases.values():
            if idb is not None:
                idb.close()

    def _lazy_load_localstorage(self):
        if self._local_storage is None:
            self._local_storage = ccl_chromium_localstorage.LocalStoreDb(self._path / LOCAL_STORAGE_FOLDER_PATH)

    def _lazy_load_sessionstorage(self):
        if self._session_storage is None:
            self._session_storage = ccl_chromium_sessionstorage.SessionStoreDb(self._path / SESSION_STORAGE_FOLDER_PATH)

    def _lazy_populate_indexeddb_list(self):
        if self._indexeddb_databases is None:
            self._indexeddb_databases = {}
            for ldb_folder in (self._path / INDEXEDDB_FOLDER_PATH).glob("*.indexeddb.leveldb"):
                if ldb_folder.is_dir():
                    idb_id = ldb_folder.name[0:-18]
                    self._indexeddb_databases[idb_id] = None

    def _lazy_load_indexeddb(self, host: str):
        self._lazy_populate_indexeddb_list()
        if host not in self._indexeddb_databases:
            raise KeyError(host)

        if self._indexeddb_databases[host] is None:
            ldb_path = self._path / INDEXEDDB_FOLDER_PATH / (host + ".indexeddb.leveldb")
            blob_path = self._path / INDEXEDDB_FOLDER_PATH / (host + ".indexeddb.blob")
            blob_path = blob_path if blob_path.exists() else None
            self._indexeddb_databases[host] = ccl_chromium_indexeddb.WrappedIndexDB(ldb_path, blob_path)

    def _lazy_load_history(self):
        if self._history is None:
            self._history = ccl_chromium_history.HistoryDatabase(self._path / HISTORY_DB_PATH)

    def _lazy_load_cache(self):
        if self._cache is None:
            if self._external_cache_folder:
                cache_path = self._external_cache_folder
            else:
                cache_path = self._path / CACHE_PATH
            cache_class = ccl_chromium_cache.guess_cache_class(cache_path)
            if cache_class is None:
                raise ValueError(f"Data under {cache_path} could not be identified as a known cache type")
            self._cache = cache_class(cache_path)

    def iter_local_storage_hosts(self) -> col_abc.Iterable[str]:
        """
        Iterates the hosts in this profile's local storage
        """
        self._lazy_load_localstorage()
        yield from self._local_storage.iter_storage_keys()

    def iter_local_storage(
            self, storage_key: typing.Optional[KeySearch]=None, script_key: typing.Optional[KeySearch]=None, *,
            include_deletions=False, raise_on_no_result=False
    ) -> col_abc.Iterable[ccl_chromium_localstorage.LocalStorageRecord]:
        """
        Iterates this profile's local storage records

        :param storage_key: storage key (host) for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool.
        :param script_key: script defined key for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool.
        :param include_deletions: if True, records related to deletions will be included
        :param raise_on_no_result: if True (the default) if no matching storage keys are found, raise a KeyError
        (these will have None as values).
        :return: iterable of LocalStorageRecords
        """
        self._lazy_load_localstorage()
        if storage_key is None and script_key is None:
            yield from self._local_storage.iter_all_records(include_deletions=include_deletions)
        elif storage_key is None:
            results = (
                x for x in self._local_storage.iter_all_records(include_deletions=include_deletions)
                if is_keysearch_hit(script_key, x.script_key))
            yielded = False
            for result in results:
                yield result
                yielded = True
            if not yielded and raise_on_no_result:
                raise KeyError(script_key)
        elif script_key is None:
            yield from self._local_storage.iter_records_for_storage_key(
                storage_key, include_deletions=include_deletions, raise_on_no_result=raise_on_no_result)
        else:
            yield from self._local_storage.iter_records_for_script_key(
                storage_key, script_key, include_deletions=include_deletions, raise_on_no_result=raise_on_no_result)

    def iter_local_storage_with_batches(
            self, storage_key: typing.Optional[KeySearch]=None, script_key: typing.Optional[KeySearch]=None, *,
            include_deletions=False, raise_on_no_result=False
    ) -> col_abc.Iterable[tuple[ccl_chromium_localstorage.LocalStorageRecord, ccl_chromium_localstorage.LocalStorageBatch]]:
        """
        Iterates this profile's local storage records with associated batches where possible.

        :param storage_key: storage key (host) for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool; or
        None (the default) in which case all hosts are considered.
        :param script_key: script defined key for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool; or
        None (the default) in which case all keys are considered.
        :param include_deletions: if True, records related to deletions will be included
        :param raise_on_no_result: if True (the default) if no matching storage keys are found, raise a KeyError
        (these will have None as values).
        :return: iterable of tuples or (LocalStorageRecords, LocalStorageBatch)
        """

        # iter_local_storage lazy loads the localstorage, so we don't need to check ahead of time.
        for rec in self.iter_local_storage(storage_key, script_key,
                                           include_deletions=include_deletions,
                                           raise_on_no_result=raise_on_no_result):
            batch = self._local_storage.find_batch(rec.leveldb_seq_number)
            yield rec, batch

    def iter_session_storage_hosts(self) -> col_abc.Iterable[str]:
        """
        Iterates this profile's session storage hosts
        """
        self._lazy_load_sessionstorage()
        yield from self._session_storage.iter_hosts()

    def iter_session_storage(
            self, host: typing.Optional[KeySearch]=None, key: typing.Optional[KeySearch]=None, *,
            include_deletions=False, raise_on_no_result=False
    ) -> col_abc.Iterable[ccl_chromium_sessionstorage.SessionStoreValue]:
        """
        Iterates this profile's session storage records

        :param host: storage key (host) for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string (each host) and
        returns a bool; or None (the default) in which case all hosts are considered.
        :param key: script defined key for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool; or
        None (the default) in which case all keys are considered.
        :param include_deletions: if True, records related to deletions will be included (these will have None as
        values).
        :param raise_on_no_result: if True, if no matching storage keys are found, raise a KeyError

        :return: iterable of SessionStoreValue
        """

        self._lazy_load_sessionstorage()
        if host is None and key is None:
            yield from self._session_storage.iter_all_records(include_deletions=include_deletions, include_orphans=True)
        elif host is None:
            results = (
                rec for
                rec in self._session_storage.iter_all_records(
                include_deletions=include_deletions, include_orphans=True)
                if is_keysearch_hit(key, rec.key)
            )
            yielded = False
            for result in results:
                yield result
                yielded = True
            if not yielded and raise_on_no_result:
                raise KeyError(key)
        elif key is None:
            yield from self._session_storage.iter_records_for_host(
                host, include_deletions=include_deletions, raise_on_no_result=raise_on_no_result)
        else:
            yield from self._session_storage.iter_records_for_session_storage_key(
                host, key, include_deletions=include_deletions, raise_on_no_result=raise_on_no_result)

    def iter_indexeddb_hosts(self) -> col_abc.Iterable[str]:
        """
        Iterates the hosts present in the Indexed DB folder. These values are what should be used to load the databases
        directly.
        """
        self._lazy_populate_indexeddb_list()
        yield from self._indexeddb_databases.keys()

    def get_indexeddb(self, host: str) -> ccl_chromium_indexeddb.WrappedIndexDB:
        """
        Returns the database with the host provided. Should be one of the values returned by
        :func:`~iter_indexeddb_hosts`. The database will be opened on-demand if it hasn't previously been opened.

        :param host: the host to get
        """
        if host not in self._indexeddb_databases:
            raise KeyError(host)

        self._lazy_load_indexeddb(host)
        return self._indexeddb_databases[host]

    def iter_indexeddb_records(
            self, host_id: typing.Optional[KeySearch], database_name: typing.Optional[KeySearch]=None,
            object_store_name: typing.Optional[KeySearch]=None, *,
            raise_on_no_result=False, include_deletions=False,
            bad_deserializer_data_handler: typing.Callable[[ccl_chromium_indexeddb.IdbKey, bytes], typing.Any] = None):
        """
        Iterates indexeddb records in this profile.

        :param host_id: the host for the records, relates to the host-named folder in the IndexedDB folder. The
        possible values for this profile are returned by :func:`~iter_indexeddb_hosts`. This can be one of:
        a single string; a collection of strings; a regex pattern; a function that takes a string (each host) and
        returns a bool; or None in which case all hosts are considered. Be cautious with supplying a parameter
        which will lead to unnecessary databases being opened as this has a set-up time for the first time it
        is opened.
        :param database_name: the database name for the records. This can be one of: a single string; a collection
        of strings; a regex pattern; a function that takes a string (each host) and returns a bool; or None (the
        default) in which case all hosts are considered.
        :param object_store_name: the object store name of the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string (each host) and returns a bool;
        or None (the default) in which case all hosts are considered.
        :param raise_on_no_result: if True, if no matching storage keys are found, raise a KeyError
        :param include_deletions: if True, records related to deletions will be included (these will have None as
        values).
        :param bad_deserializer_data_handler: a callback function which will be executed by the underlying
        indexeddb reader if invalid data is encountered during reading a record, rather than raising an exception.
        The function should take two arguments: an IdbKey object (which is the key of the bad record) and a bytes
        object (which is the raw data). The return value of the callback is ignored by the calling code. If this is
        None (the default) then any bad data will cause an exception to be raised.

        """
        self._lazy_populate_indexeddb_list()

        # probably not optimal performance, but we only do it once per call, and it's a lot neater.
        if host_id is None:
            found_hosts = list(self.iter_indexeddb_hosts())
        else:
            found_hosts = [x for x in self._indexeddb_databases.keys() if is_keysearch_hit(host_id, x)]

        if not found_hosts and raise_on_no_result:
            raise KeyError((host_id, database_name, object_store_name))

        yielded = False
        for found_host in found_hosts:
            idb = self.get_indexeddb(found_host)
            for idb_db_id in idb.database_ids:
                if database_name is None or is_keysearch_hit(database_name, idb_db_id.name):
                    idb_db = idb[idb_db_id]
                    for idb_db_objstore_name in idb_db.object_store_names:
                        if object_store_name is None or is_keysearch_hit(object_store_name,  idb_db_objstore_name):
                            idb_db_objstore = idb_db.get_object_store_by_name(idb_db_objstore_name)
                            for rec in idb_db_objstore.iterate_records(
                                    live_only=not include_deletions,
                                    bad_deserializer_data_handler=bad_deserializer_data_handler):
                                yield rec
                                yielded = True

        if not yielded and raise_on_no_result:
            raise KeyError((host_id, database_name, object_store_name))

    def iterate_history_records(
            self, url: typing.Optional[KeySearch]=None, *,
            earliest: typing.Optional[datetime.datetime]=None, latest: typing.Optional[datetime.datetime]=None):
        """
        Iterates history records for this profile.

        :param url: a URL to search for. This can be one of: a single string; a collection of strings;
        a regex pattern; a function that takes a string (each host) and returns a bool; or None (the
        default) in which case all hosts are considered.
        :param earliest: an optional datetime which will be used to exclude records before this date.
        NB the date should be UTC to match the database. If None, no lower limit will be placed on
        timestamps.
        :param latest: an optional datetime which will be used to exclude records after this date.
        NB the date should be UTC to match the database. If None, no upper limit will be placed on
        timestamps.
        """
        self._lazy_load_history()
        yield from self._history.iter_history_records(url, earliest=earliest, latest=latest)

    @staticmethod
    def _decompress_cache_data(data, content_encoding) -> tuple[bool, bytes]:
        try:
            if content_encoding.strip() == "gzip":
                return True, gzip.decompress(data)
            elif content_encoding.strip() == "br":
                return True, brotli.decompress(data)
            elif content_encoding.strip() == "deflate":
                return True, zlib.decompress(data, -zlib.MAX_WBITS)  # suppress trying to read a header
        except (EOFError, gzip.BadGzipFile, brotli.error, zlib.error):
            return False, data

        return False, data

    def _yield_cache_record(self, key: ccl_chromium_cache.CacheKey, decompress, omit_data):
        metas = self._cache.get_metadata(key)
        if not omit_data:
            datas = self._cache.get_cachefile(key)
        else:
            datas = [None] * len(metas)
        meta_locations = self._cache.get_location_for_metadata(key)
        data_locations = self._cache.get_location_for_cachefile(key)

        if not (len(metas) == len(datas) == len(meta_locations) == len(data_locations)):
            raise ValueError("Data and metadata counts do not match")

        for idx, (meta, data, meta_location, data_location) in enumerate(
                zip(metas, datas, meta_locations, data_locations)):
            if decompress and data is not None and meta is not None:
                content_encoding = (meta.get_attribute("content-encoding") or [""])[0]
                was_decompressed, data = self._decompress_cache_data(data, content_encoding)
            else:
                was_decompressed = False
            yield CacheResult(key, meta, data, meta_location, data_location, was_decompressed, idx)

    def iterate_cache(
            self,
            url: typing.Optional[KeySearch]=None, *, decompress=True, omit_cached_data=False,
            **kwargs: typing.Union[bool, KeySearch]) -> col_abc.Iterable[CacheResult]:
        """
        Iterates cache records for this profile.

        :param url: a URL to search for. This can be one of: a single string; a collection of strings;
        a regex pattern; a function that takes a string (each host) and returns a bool; or None (the
        default) in which case all records are considered.
        :param decompress: if True (the default), data from the cache which is compressed (as per the
        content-encoding header field) will be decompressed when read if the compression format is
        supported (currently deflate, gzip and brotli are supported).
        :param omit_cached_data: does not collect the cached data and omits it from each `CacheResult`
        object. Should be faster in cases when only metadata recovery is required.
        :param kwargs: further keyword arguments are used to search based upon header fields. The
        keyword should be the header field name, with underscores replacing hyphens (e.g.,
        content-encoding, becomes content_encoding). The value should be one of: a Boolean (in which
        case only records with this field present will be included if True, and vice versa); a single
        string; a collection of strings; a regex pattern; a function that takes a string (the value)
        and returns a bool.
        """

        self._lazy_load_cache()
        if url is None and not kwargs:
            for key in self._cache.cache_keys():
                yield from self._yield_cache_record(key, decompress, omit_cached_data)
        else:
            for key in self._cache.cache_keys():
                if url is not None and not is_keysearch_hit(url, key.url):
                    # Fail condition: URL doesn't match
                    continue
                if not kwargs:
                    # No metadata keyword arguments to check
                    yield from self._yield_cache_record(key, decompress, omit_cached_data)
                else:
                    metas = self._cache.get_metadata(key)
                    if not metas or all(x is None for x in metas):
                        # Fail condition: we had metadata to check, but no metadata to check against
                        continue
                    else:
                        meta_hit_indices = []
                        for meta_idx, meta in enumerate(metas):
                            hit = True
                            for attribute_name, attribute_check in kwargs.items():
                                attribute_name = attribute_name.replace("_", "-")
                                if isinstance(attribute_check, bool):
                                    if (attribute_check == True and
                                            not meta.has_declaration(attribute_name) and
                                            not meta.get_attribute(attribute_name)):
                                        hit = False
                                        break
                                    if (attribute_check == False and
                                            (meta.has_declaration(attribute_name) or
                                             meta.get_attribute(attribute_name))):
                                        hit = False
                                        break
                                else:
                                    attribute = meta.get_attribute(attribute_name)
                                    if not any(is_keysearch_hit(attribute_check, x) for x in attribute):
                                        hit = False
                                        break

                            if hit:
                                meta_hit_indices.append(meta_idx)

                        if meta_hit_indices:
                            if not omit_cached_data:
                                datas = self._cache.get_cachefile(key)
                            else:
                                datas = [None] * len(metas)
                            metadata_locations = self._cache.get_location_for_metadata(key)
                            data_locations = self._cache.get_location_for_cachefile(key)

                            if not (len(metas) == len(datas) == len(metadata_locations) == len(data_locations)):
                                raise ValueError("Data and metadata counts do not match")

                            for i in meta_hit_indices:
                                meta = metas[i]
                                data = datas[i]
                                metadata_location = metadata_locations[i]
                                data_location = data_locations[i]

                                if decompress and data is not None:
                                    content_encoding = (meta.get_attribute("content-encoding") or [""])[0]
                                    was_decompressed, data = self._decompress_cache_data(data, content_encoding)
                                else:
                                    was_decompressed = False

                                yield CacheResult(
                                    key, meta, data, metadata_location, data_location, was_decompressed, i)

    def iter_downloads(
            self, *, download_url: typing.Optional[KeySearch]=None, tab_url: typing.Optional[KeySearch]=None):
        """
        Iterates download records for this profile

        :param download_url: A URL related to the downloaded resource. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string (each host) and returns a bool;
        or None (the default) in which case all records are considered.
        :param tab_url: A URL related to the page the user was accessing when this download was started.
        This can be one of: a single string; a collection of strings; a regex pattern; a function that takes
        a string (each host) and returns a bool; or None (the default) in which case all records are considered.
        """
        for download in ccl_shared_proto_db_downloads.read_downloads(self._path / SHARED_PROTO_DB_FOLDER_PATH):
            if ((download_url is None or any(is_keysearch_hit(download_url, url) for url in download.url_chain))
                and
                (tab_url is None
                    or is_keysearch_hit(tab_url, download.tab_url or "")
                    or is_keysearch_hit(tab_url, download.tab_referrer_url or ""))):
                yield download

        self._lazy_load_history()
        for download in self._history.iter_downloads(download_url=download_url, tab_url=tab_url):
            yield download

    def __enter__(self) -> "ChromiumProfileFolder":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def path(self):
        """The input path of this profile folder"""
        return self._path

    @property
    def local_storage(self) -> ccl_chromium_localstorage.LocalStoreDb:
        """The local storage object for this profile folder"""
        self._lazy_load_localstorage()
        return self._local_storage

    @property
    def session_storage(self) -> ccl_chromium_sessionstorage.SessionStoreDb:
        """The session storage object for this profile folder"""
        self._lazy_load_sessionstorage()
        return self._session_storage

    @property
    def cache(self) -> ccl_chromium_cache.ChromiumCache:
        """The cache for this profile folder"""
        self._lazy_load_cache()
        return self._cache

    @property
    def history(self) -> ccl_chromium_history.HistoryDatabase:
        """The history for this profile folder"""
        self._lazy_load_history()
        return self._history

    @property
    def browser_type(self) -> str:
        return "Chromium"



import dataclasses
import datetime
import struct
import enum

from .serialization_formats import ccl_protobuff as pb


class DownloadSource(enum.Enum):
    shared_proto_db = 1
    history_db = 2


@dataclasses.dataclass(frozen=True)
class Download:  # TODO: all of the parameters
    record_source: DownloadSource
    record_id: int
    guid: str
    hash: str
    url_chain: tuple[str, ...]
    tab_url: str
    tab_referrer_url: str
    target_path: str
    mime_type: str
    original_mime_type: str
    total_bytes: str
    start_time: datetime.datetime
    end_time: datetime.datetime

    @property
    def level_db_seq_no(self):
        if self.record_source == DownloadSource.shared_proto_db:
            return self.record_id

    @property
    def record_location(self) -> str:
        if self.record_source == DownloadSource.shared_proto_db:
            return f"Leveldb Seq: {self.record_id}"
        elif self.record_source == DownloadSource.history_db:
            return f"SQLite Rowid: {self.record_id}"
        raise NotImplementedError()

    @property
    def url(self) -> str:
        return self.url_chain[-1]

    @property
    def file_size(self) -> int:
        return int(self.total_bytes)

    @classmethod
    def from_pb(cls, seq: int, proto: pb.ProtoObject, *, target_path_is_utf_16=True):
        if not proto.only("download_info").value:
            raise ValueError("download_info is empty")
        target_path_raw = proto.only("download_info").only("in_progress_info").only("target_path").value
        path_proto_length, path_char_count = struct.unpack("<II", target_path_raw[0:8])
        if path_proto_length != len(target_path_raw) - 4:
            raise ValueError("Invalid pickle for target path")
        if target_path_is_utf_16:
            target_path = target_path_raw[8: 8 + (path_char_count * 2)].decode("utf-16-le")
        else:
            target_path = target_path_raw[8: 8 + path_char_count].decode("utf-8")

        return cls(
            DownloadSource.shared_proto_db,
            seq,
            proto.only("download_info").only("guid").value,
            proto.only("download_info").only("in_progress_info").only("hash").value.hex(),
            tuple(x.value for x in proto.only("download_info").only("in_progress_info")["url_chain"]),
            proto.only("download_info").only("in_progress_info").only("tab_url").value,
            proto.only("download_info").only("in_progress_info").only("tab_url_referrer").value,
            target_path,
            proto.only("download_info").only("in_progress_info").only("mime_type").value,
            proto.only("download_info").only("in_progress_info").only("original_mime_type").value,
            proto.only("download_info").only("in_progress_info").only("total_bytes").value,
            proto.only("download_info").only("in_progress_info").only("start_time").value,
            proto.only("download_info").only("in_progress_info").only("end_time").value,
        )



from __future__ import annotations

from typing import Any

import pytest

from linear_mcp_fast import local_handlers
from linear_mcp_fast.official_session import OfficialToolError
from linear_mcp_fast.router import ToolRouter


class FakeReader:
    def __init__(self, degraded: bool = False):
        self.degraded = degraded
        self.refresh_count = 0

    def is_degraded(self) -> bool:
        return self.degraded

    def refresh_cache(self, force: bool = True) -> None:
        self.refresh_count += 1

    def get_health(self) -> dict[str, Any]:
        return {"degraded": self.degraded, "refreshCount": self.refresh_count}


class FakeOfficial:
    def __init__(self):
        self.calls: list[tuple[str, dict[str, Any]]] = []
        self.responses: dict[str, Any] = {}
        self.exceptions: dict[str, Exception] = {}

    def call_tool(self, name: str, arguments: dict[str, Any] | None = None) -> Any:
        args = arguments or {}
        self.calls.append((name, args))
        if name in self.exceptions:
            raise self.exceptions[name]
        return self.responses.get(name, {"ok": True, "tool": name, "args": args})

    def list_tools(self) -> list[str]:
        return ["create_issue", "list_issues"]

    def get_health(self) -> dict[str, Any]:
        return {"connected": True}


def _install_local_handler(monkeypatch: pytest.MonkeyPatch, fn):
    monkeypatch.setitem(local_handlers.LOCAL_READ_HANDLERS, "list_issues", fn)


def test_read_local_success_without_official_call(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    result = router.call_read("list_issues", {})

    assert result == {"source": "local"}
    assert official.calls == []


def test_read_local_unsupported_falls_back_to_official(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()

    def handler(_reader, **_kwargs):
        raise local_handlers.LocalFallbackRequested("unsupported_filter", "unsupported")

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    result = router.call_read("list_issues", {"query": "hello"})

    assert result["tool"] == "list_issues"
    assert official.calls == [("list_issues", {"query": "hello"})]


def test_degraded_local_uses_official_when_available(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=True)
    official = FakeOfficial()

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    result = router.call_read("list_issues", {})

    assert result["tool"] == "list_issues"
    assert official.calls == [("list_issues", {})]


def test_degraded_local_returns_stale_local_if_remote_down(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=True)
    official = FakeOfficial()
    official.exceptions["list_issues"] = OfficialToolError("official_down", "offline")

    def handler(_reader, **_kwargs):
        return {"source": "local-stale"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    result = router.call_read("list_issues", {})

    assert result == {"source": "local-stale"}
    assert len(official.calls) == 1


def test_unexpected_local_error_falls_back_to_official(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()

    def handler(_reader, **_kwargs):
        raise RuntimeError("boom")

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    result = router.call_read("list_issues", {})

    assert result["tool"] == "list_issues"
    assert len(official.calls) == 1


def test_write_marks_coherence_and_uses_remote_first(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()
    official.responses["create_issue"] = {"id": "ISS-1"}
    official.responses["list_issues"] = {"source": "remote"}

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    router.call_official("create_issue", {"title": "T"})
    result = router.call_read("list_issues", {})

    assert result == {"source": "remote"}
    assert official.calls[0][0] == "create_issue"
    assert official.calls[1][0] == "list_issues"


def test_remote_first_falls_back_to_local_when_remote_fails(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()
    official.responses["create_issue"] = {"id": "ISS-1"}
    official.exceptions["list_issues"] = OfficialToolError("official_down", "offline")

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    router.call_official("create_issue", {"title": "T"})
    result = router.call_read("list_issues", {})

    assert result == {"source": "local"}
    assert len(official.calls) == 2


def test_remote_first_tool_error_does_not_fallback_to_local(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()
    official.responses["create_issue"] = {"id": "ISS-1"}
    official.exceptions["list_issues"] = OfficialToolError("official_tool_error", "bad args")

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    router.call_official("create_issue", {"title": "T"})
    with pytest.raises(OfficialToolError) as exc_info:
        router.call_read("list_issues", {})

    assert exc_info.value.code == "official_tool_error"
    assert len(official.calls) == 2


def test_remote_first_degraded_local_does_not_retry_official_twice(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=True)
    official = FakeOfficial()
    official.responses["create_issue"] = {"id": "ISS-1"}
    official.exceptions["list_issues"] = OfficialToolError("official_unavailable", "offline")

    def handler(_reader, **_kwargs):
        return {"source": "local-stale"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    router.call_official("create_issue", {"title": "T"})
    result = router.call_read("list_issues", {})

    assert result == {"source": "local-stale"}
    assert len(official.calls) == 2


def test_non_write_official_call_does_not_force_remote_first(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    router.call_official("list_teams", {})
    result = router.call_read("list_issues", {})

    assert result == {"source": "local"}


def test_refresh_local_cache_returns_local_health(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    health = router.refresh_local_cache()

    assert health["refreshCount"] == 1


def test_router_health_includes_local_and_official(monkeypatch: pytest.MonkeyPatch):
    reader = FakeReader(degraded=False)
    official = FakeOfficial()

    def handler(_reader, **_kwargs):
        return {"source": "local"}

    _install_local_handler(monkeypatch, handler)

    router = ToolRouter(reader, official, coherence_window_seconds=30)
    health = router.get_health()

    assert "local" in health
    assert "official" in health
    assert health["coherenceWindowSeconds"] == 30
